/*
 *
 * Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002 Yuuji ICHISUGI
 *
 * Permission to use, copy, modify and redistribution this software in
 * whole and in part, for evaluation or research purposes and without fee
 * is hereby granted provided this copyright notice.
 * See CopyrightAndLicensing.txt for licensing condition.
 */

/*
 NOTE: This plug-in requires "AutoSplitFiles" and "Symbol" plug-in.
*/

#epp Symbol
#epp SystemMixin
#epp AutoSplitFiles
#epp BackQuote

package SystemMixin;
import epp.*;
import SystemMixinRuntimes.*;
import java.util.Vector;
import java.util.Hashtable;


SystemMixin SystemMixin {
  class Epp {
    extend void initExtraImportDeclarations(){
      original();
      /* old
      addExtraImportDeclaration(`(importOnDemand
				  (name
				   (id SystemMixinRuntimes))));
      */
    }
    
    extend Tree typeDeclarationTop(){
      if (lookahead() == :SystemMixin){
	matchAny();
	TreeVec tvec = new TreeVec();
	tvec.add(identifier());
	match(:"{");
	while (lookahead() != :"}"){
	  tvec.add(classMixin());
	}
	matchAny();
	return new Tree(:systemMixin, tvec.toArray());
      } else if (lookahead() == :BeforeLoadingPlugIn){
	matchAny();
	return new Tree(:beforeLoadingPlugIn, block());
      } else {
	return original();
      }
    }

    define implement Tree classMixin(){
      return typeDeclaration();
    }
    
    extend Tree modifierTop() {
      Token token = lookahead();
      if (token == :define ||
	  token == :default || token == :implement ||
	  token == :extend || token == :redefine) {
	return identifier();
      } else {
	return original();
      }
    }

    extend Tree primaryTop() {
      if (lookahead() == :original){
	matchAny();
	//old match(:"(");
	Tree args = argumentList();
	//old match(:")");
	return new Tree(:original, args);
      } else {
	return original();
      }
    }

    extend void beforeMacroExpansionPass(){
      String inputFileName = (String)Dynamic.get(:inputFileName);
      Symbol loaderName
	= Symbol.intern(Epp.getClassNameFromFileName(inputFileName));

      Tree[] typeDecls = self!tree.args()[2].args();
      for (int i = 0; i < typeDecls.length; i++){
	if (typeDecls[i].tag() == :class){
	  Symbol className = typeDecls[i].args()[2].idName();
	  if (className == loaderName){
	    throw error("The class name "+ className+
			" collides with PlugInLoder name "+
			"generated by SystemMixin plug-in. Please rename it.");
	  }
	}
      }
    }

    extend void initMacroTable() {
      original();
      extendMacro(:typeDeclarations, new SystemMixinTypeDeclarationsMacro());
      defineMacro(:systemMixin, new SystemMixinMacro());
      defineMacro(:beforeLoadingPlugIn, new BeforeLoadingPlugInMacro());
      defineMacro(:original, new OriginalMacro());
      defineMacro(:this, new ThisMacro());
    }
  }
}



class SystemMixinTypeDeclarationsMacro extends Macro {
  
  public Tree call(Tree tree){
    Dynamic.bind(:systemMixinsInThisFile, new TreeVec());
    Dynamic.bind(:beforeLoadingPlugInInThisFile, new TreeVec());
    Dynamic.bind(:expandingSystemName, null);
    Dynamic.bind(:expandingClassName, null);
    Dynamic.bind(:expandingMethodName, null);
    try{
      
      // Expands SystemMixin declaratoins and class declarations.
      Tree etree = orig.call(tree);
      
      String inputFileName = (String)Dynamic.get(:inputFileName);
      
      Tree loaderName
	= new Identifier(Symbol.intern
			 (Epp.getClassNameFromFileName(inputFileName)));
      
      TreeVec tvec = new TreeVec();
      Tree[] systemMixins =
	((TreeVec)Dynamic.get(:systemMixinsInThisFile)).toArray();
      for (int i = 0; i < systemMixins.length; i++){
	// (fieldLong (name SystemMixin_s1) (id instance))
	tvec.add(
		 `(fieldLong
		   (name ,(new Identifier(systemMixins[i].idName())))
		   (id instance))
		 );
      }
      return
	`(typeDeclarations
	  ,(macroExpandTree
	    (
	     `(class
	       (modifiers (id public))
	       (id class)
	       ,loaderName
	       (extends (name (id PlugInLoader)))
	       (implements)
	       (classBody
		(constructor
		 (modifiers (id public)) (name (id void))
		 ,loaderName (formalParameters)
		 (throws)
		 (block (decl
			 (modifiers)
			 (arrayOf (name (id SystemMixin)))
			 (vardecls
			  (varInit (id ma)
			   (arrayInitializer
			    ,@tvec
			    ))))
		  (expressionStatement (= (id mixins) (id ma)))))
		(method
		 (modifiers (id public) (id static)) (name (id void)) (id main)
		 (formalParameters (pdecl (modifiers)
				    (arrayOf (name (id String))) (id args)))
		 (throws)
		 (block
		  (expressionStatement
		   (invokeLong (name (id PlugInLoader))
		    (id doMain)
		    (argumentList
		     (new (name ,loaderName) (argumentList))
		     (id args))))))
		(method
		 (modifiers (id public))
		 (name (id void))
		 (id beforeLoadingPlugIn) (formalParameters) (throws)
		 (block
		  ,@((TreeVec)Dynamic.get(:beforeLoadingPlugInInThisFile))))
		))))
	  ,@(etree.args())
	  );
    } finally {
      Dynamic.unbind();
      Dynamic.unbind();
      Dynamic.unbind();
      Dynamic.unbind();
      Dynamic.unbind();
    }
  }
}



class BeforeLoadingPlugInMacro extends Macro {
  public Tree call(Tree tree){
    Tree block = tree.args()[0];
    ((TreeVec)Dynamic.get(:beforeLoadingPlugInInThisFile)).add(block);
    return `(emptyTypeDeclaration);
  }
}



class SystemMixinMacro extends Macro {
  public Tree call(Tree tree){
    Tree[] args			= tree.args();
    String name			= args[0].idName().toString();
    TreeVec methodMixinLists	= new TreeVec();
    TreeVec classMixins		= new TreeVec();
    
    Dynamic.bind(:expandingSystemName, name);
    try{
      for (int i = 1; i < args.length; i++){
	expandClassMixin(args[i], methodMixinLists, classMixins);
      }
      
      ((TreeVec)Dynamic.get(:systemMixinsInThisFile))
        .add(Util.mixinNameToClass(name));
      
      addTree(:beforeCurrentTypeDeclaration,
	      `(class
		(modifiers)
		(id class)
		,(Util.mixinNameToClass(name))
		(extends (name (id SystemMixin)))
		(implements)
		(classBody
		 (decl (modifiers (id public) (id static))
		  (name ,(Util.mixinNameToClass(name)))
		  (vardecls
		   (varInit (id instance)
		    (new (name ,(Util.mixinNameToClass(name)))
		     (argumentList)))))
		 (constructor (modifiers (id private)) (name (id void))
		  ,(Util.mixinNameToClass(name))
		  (formalParameters) (throws)
		  (block (expressionStatement
			  (= (id name)
			   ,(Util.nameToSymbol(Util.getPackageName()+
					       name))))
		   ,@methodMixinLists
		   (decl (modifiers) (arrayOf (name (id ClassMixin)))
		    (vardecls
		     (varInit (id cm)
		      (arrayInitializer
		       ,@classMixins))))
		   (expressionStatement (= (id classMixins) (id cm)))
		   (expressionStatement
		    (= (id lightFlag)
		     ,(Opts.sml ? `(id true) : `(id false))))
		   ))))
	      );
      
      return `(emptyTypeDeclaration);
    } finally {
      Dynamic.unbind();
    }
  }
  
  
  void expandClassMixin(Tree tree,
			TreeVec methodMixinLists, TreeVec classMixins){
    checkArgsLength(tree, 6);
    Tree[] args			= tree.args();
    Tree modifiers		= args[0];
    Tree keyword		= args[1];
    String name			= args[2].idName().toString();
    Tree extendsTree		= args[3];
    Tree implementsTree		= args[4];
    Tree classBody		= args[5];
    
    Dynamic.bind(:expandingClassName, name);
    try{
      Tree methodMixinsTmpName	= new Identifier(Symbol.intern("mms_"+ name));
      TreeVec methodMixins	= new TreeVec();
      TreeVec fields		= new TreeVec();
      Tree[] supers 		= extendsTree.args();
      Tree fieldAllocatorClassName
	= new Identifier(Symbol.intern("ObjFieldAllocator_"+
				       (String)Dynamic.get(:expandingSystemName)+
				       "_"+
				       (String)Dynamic.get(:expandingClassName)));
      
      boolean defineIsSpecified = checkIfDefineIsSpecified(modifiers);
      
      if (supers.length > 1) {
	throw systemMixinUserError("Multiple inheritance is not supported.");
      }
      
      // Start collecting constructor, method, field declarations.
      Tree constructorArgs = null;
      {
	Tree[] classBodyArgs = classBody.args();
	for (int i = 0; i < classBodyArgs.length; i++){
	  Tree decl = classBodyArgs[i];
	  Symbol tag = decl.tag();
	  
	  if (tag == :method) {
	    methodMixins.add(expandMethodMixin(decl));
	    if (defineIsSpecified){
	      Tree ret = checkConstructor(decl, args[2].idName());
	      if (ret != null && constructorArgs != null){
		throw systemMixinUserError("Multiple constructors are declared.");
	      }
	      if (ret != null) {
		constructorArgs = ret;
	      }
	    }
	  } else if (tag == :constructor) {
	    throw systemMixinUserError
	      ("A result type \"void\" should be specified to the constructor.");
	  } else if (tag == :decl) {
	    fields.add(expandFieldMixin(decl));
	  } else {
	    throw systemMixinUserError("FATAL");
	  }
	}
      }
      
      if (defineIsSpecified){
	if (constructorArgs == null){
	  throw systemMixinUserError("There are no constructor.");
	} else {
	  addTree(:beforeCurrentTypeDeclaration,
		  `(class
		    (modifiers (id public))
		    (id class)
		    ,(Util.classNameToClass(name))
		    (extends)
		    (implements)
		    (classBody
		     (method
		      (modifiers (id public) (id static))
		      (name (id Obj))
		      (id generateAndInit)
		      ,constructorArgs
		      (throws)
		      (block (decl (modifiers) (name (id Obj))
			      (vardecls
			       (varInit (id obj)
				(invokeLong
				 (name (id SystemMixin))
				 (id generate)
				 (argumentList
				  ,(Util.nameToSymbol(name)))))))
		       (expressionStatement
			(invokeExp
			 (cast (name ,(Util.methodNameToClass(name)))
			  (invokeLong (name (id obj))
			   (id findMethod)
			   (argumentList
			    ,(Util.nameToSymbol(name)))))
			 (id call)
			 ,(Opts.sml ? `(argumentList
					,@(fparamsToAparams(constructorArgs)))
			   : `(argumentList
			       (fieldLong (name (id Obj))
				(id calleeself))
			       (fieldLong (name (id Obj))
				(id calleemb))
			       (fieldLong (name (id Obj))
				(id calleembi))
			       ,@(fparamsToAparams(constructorArgs))))
			 ))
		       (returnWithExp (id obj))))))
		  );
	}
      }
      
      addTree(:beforeCurrentTypeDeclaration,
	      `(class
		(modifiers)
		(id class)
		,fieldAllocatorClassName
		(extends (name (id ObjFieldAllocator)))
		(implements)
		(classBody
		 (method
		  (modifiers (id public))
		  (name (id void))
		  (id allocate)
		  (formalParameters
		   (pdecl (modifiers)
		    (name (id java) (id util) (id Hashtable))
		    (id table)))
		  (throws)
		  (block ,@fields))))
	      );
      
      methodMixinLists.add(
			   `(decl (modifiers) (arrayOf (name (id MethodMixin)))
			     (vardecls
			      (varInit ,methodMixinsTmpName
			       (arrayInitializer
				,@methodMixins))))
			   );
      classMixins.add(
		      `(new (name (id ClassMixin))
			(argumentList
			 ,(Util.nameToSymbol(name))
			 ,(defineIsSpecified ? `(id true) : `(id false))
			 ,(supers.length == 1 ?
			   // Contents of supers[0] is (name (id "superName")).
			   Util.nameToSymbol(supers[0].args()[0].idName().toString())
			   : `(id null))
			 (new
			  (name
			   ,fieldAllocatorClassName)
			  (argumentList))
			 ,methodMixinsTmpName))
		      );
    } finally {
      Dynamic.unbind();
    }
  }
  
  
  Tree checkConstructor(Tree tree, Symbol className){
    Tree[] args = tree.args();
    Symbol name = args[2].idName();
    if (name == className){
      Tree modifiers	= args[0];
      Tree result	= args[1];
      Tree fparams	= args[3];
      boolean defineIsSpecified = checkIfDefineIsSpecified(modifiers);
      
      if (! defineIsSpecified){
	throw systemMixinUserError
	  ("The constructor is not \"define\" method.");
      }
      if (! (result.tag() == :name &&
	     result.args()[0].idName() == :void)) {
	throw systemMixinUserError
	  ("The result type of the constructor is not void.");
      }
      return fparams;
    }
    return null;
  }
  
  Tree[] fparamsToAparams(Tree constructorArgs){
    //    (pdecl (modifiers) (arrayOf (name (id String)))
    //           (id argv))
    // -> (id argv)
    Tree[] fparams = constructorArgs.args();
    Tree[] aparams = new Tree[fparams.length];
    for (int i = 0; i < fparams.length; i++){
      Tree var =  fparams[i].args()[2];
      if (var.tag() != :id){
	throw systemMixinUserError("Declaration style like foo(int a[]) is not supported.  Please use foo(int[] a) .");
      }
      aparams[i] = var;
    }
    return aparams;
  }
  
  
  boolean checkIfDefineIsSpecified(Tree modifiers){
    boolean defineIsSpecified = false;
    Tree[] mods = modifiers.args();
    for (int i = 0; i < mods.length; i++){
      if (mods[i].idName() == :define) {
	return true;
      }
    }
    return false;
  }
  
  
  Tree expandMethodMixin(Tree tree){
    checkArgsLength(tree, 6);
    Tree[] args			= tree.args();
    Tree modifiers		= args[0];
    Tree result			= args[1];
    String name			= args[2].idName().toString();
    Tree fparams		= args[3];
    Tree throwsTree		= args[4];
    Tree methodBody		= args[5];
    
    Dynamic.bind(:expandingMethodName, name);
    try{
      boolean defineIsSpecified = checkIfDefineIsSpecified(modifiers);
      Symbol methodRole = checkMethodRole(modifiers);
      Tree methodBodyClassName
	= new Identifier(Symbol.intern((String)Dynamic.get(:expandingSystemName)+
				       "_"+
				       (String)Dynamic.get(:expandingClassName)+
				       "_"+
				       name));
      
      if (methodRole == :abstract){
	if (methodBody.tag() != :emptyStatement){
	  throw systemMixinUserError
	    ("The method body of this abstract method is not emptyStatement.");
	}
      }
      
      if (defineIsSpecified){
	addTree(:beforeCurrentTypeDeclaration,
		`(class
		  (modifiers (id public) (id abstract))
		  (id class)
		  ,(Util.methodNameToClass(name))
		  (extends)
		  (implements)
		  (classBody
		   (decl (modifiers (id public) (id static))
		    (name ,(Util.methodNameToClass(name)))
		    (vardecls (id m)))
		   (decl (modifiers (id public))
		    (name ,(Util.methodNameToClass(name)))
		    (vardecls (id _original)))
		   ,@(Opts.sml ?
		      new Tree[]{`(decl (modifiers (id public) (id static))
				   (name (id Obj))
				   (vardecls (id self)))}
		      : new Tree[0] )
		   (method
		    (modifiers (id public) (id abstract))
		    ,result
		    (id call)
		    ,(Opts.sml ? `(formalParameters ,@(fparams.args()))
		      : `(formalParameters
			  (pdecl (modifiers) (name (id Obj)) (id self))
			  (pdecl (modifiers) (arrayOf (name (id Object)))
			   (id _mb))
			  (pdecl (modifiers) (name (id int)) (id _mbi))
			  ,@(fparams.args())))
		    ,throwsTree
		    (emptyStatement))))
		);
      }
      
      addTree(:beforeCurrentTypeDeclaration,
	      `(class
		(modifiers)
		(id class)
		,methodBodyClassName
		(extends (name ,(Util.methodNameToClass(name))))
		(implements)
		(classBody
		 (method
		  (modifiers (id public))
		  ,result
		  (id call)
		  ,(Opts.sml ? `(formalParameters ,@(fparams.args()))
		    : `(formalParameters
			(pdecl (modifiers) (name (id Obj)) (id self))
			(pdecl (modifiers) (arrayOf (name (id Object)))
			 (id _mb))
			(pdecl (modifiers) (name (id int)) (id _mbi))
			,@(fparams.args())))
		  ,throwsTree
		  ,(methodRole == :abstract ?
		    `(block
		      (throw (new (name (id Error))
			      (argumentList
			       (literalTree
				string
				"FATAL: Abstract LdMethod is called.")))))
		    : methodBody)
		  )))
	      );
      
      return `(new (name (id MethodMixin))
	       (argumentList ,(Util.nameToSymbol(name))
		,(defineIsSpecified ? `(id true) : `(id false))
		,(Util.nameToSymbol(methodRole.toString()))
		(new (name ,methodBodyClassName) (argumentList))));
    } finally {
      Dynamic.unbind();
    }
  }
  
  
  static final Symbol methodRoleList[] =
    {:abstract, :default, :implement, :extend, :redefine};
  
  Symbol checkMethodRole(Tree modifiers){
    Tree[] mods = modifiers.args();
    Symbol methodRole = :unspecified; // default method role
    for (int i = 0; i < mods.length; i++){
      Symbol mod = mods[i].idName();
      if (mod == :define) {
	// Do nothing.
      } else {
      checkRoleList:
	{
	  for (int j = 0; j < methodRoleList.length; j++){
	    if (mod == methodRoleList[j]){
	      if (methodRole == :unspecified){
		methodRole = mod;
		break checkRoleList;
	      } else {
		throw systemMixinUserError
		  ("More than one method roles are specified: "+
		   methodRole+
		   ", "+
		   mod+
		   ", ...");
	      }
	    }
	  }
	  throw systemMixinUserError("Illegal modifier are specified: "+ mod);
	}
      }
    }
    return methodRole;
  }
  

  Tree expandFieldMixin(Tree tree){
    // Expected tree is:
    //  (decl (modifiers) (name (id int)) (vardecls (id v1)))
    Tree[] args = tree.args();
    Tree modifiers		= args[0]; // This value is ignored.
    Tree type			= args[1];
    Tree vardecls		= args[2];
    
    if (vardecls.tag() != :vardecls) {
      throw systemMixinUserError("FATAL");
    }
    Tree[] vardeclsArgs = vardecls.args();
    if (! (vardeclsArgs.length == 1
	   && vardeclsArgs[0].tag() == :id)){
      throw systemMixinUserError("Field initialization is not supported.");
    }
    String name = vardeclsArgs[0].idName().toString();

    addTree(:beforeCurrentTypeDeclaration,
	    `(class
	      (modifiers (id public))
	      (id class)
	      ,(Util.fieldNameToClass(name))
	      (extends)
	      (implements)
	      (classBody
	       (decl (modifiers (id public) (id static))
		,type (vardecls (id v)))
	       (decl (modifiers (id public)) ,type (vardecls (id value)))))
	    );
    
    return `(expressionStatement
	     (invokeLong (name (id table)) (id put)
	      (argumentList ,(Util.nameToSymbol(name))
	       (new (name ,(Util.fieldNameToClass(name))) (argumentList)))));
  }
  
  
  Error systemMixinUserError(String str){
    return error("Expanding systemMixin : "+
		 Dynamic.get(:expandingSystemName)+
		 "/"+
		 Dynamic.get(:expandingClassName)+
		 "/"+
		 Dynamic.get(:expandingMethodName)+
		 " : "+
		 str);
  }
}



class OriginalMacro extends Macro {
  public Tree call(Tree tree){
    checkArgsLength(tree, 1);
    Tree[] args = tree.args();
    Tree aparams = args[0];
    if (Dynamic.get(:expandingSystemName) == null){
      throw error("\"original\" appeared in a normal Java method.");
    }
    if (Opts.sml){
      // _original.call(a, b)
      return `(invokeLong (name (id _original))
	       (id call)
	       (argumentList ,@(aparams.args())));
    } else {
      // ((M_foo)_mb[_mbi]).call(self, _mb, _mbi - 1, a, b)
      return `(invokeExp
	       (cast (name ,(Util.methodNameToClass
			     ((String)Dynamic.get(:expandingMethodName))))
		(array (id _mb) (id _mbi)))
	       (id call)
	       (argumentList (id self) (id _mb)
		(- (id _mbi) (literalTree int "1"))
		,@(aparams.args())));
    }
  }
}



class ThisMacro extends Macro {
  public Tree call(Tree tree){
    checkArgsLength(tree, 0);
    if (Dynamic.get(:expandingSystemName) != null){
      throw error("\"original\" appeared in a normal Java method.");
    }
    return tree;
  }
}


//-------------------------------------------------
//--------------------------------------------------

SystemMixin UseLdObj {
  class Epp {
    extend Tree primaryLeft(Tree tree){
      // obj!x or obj!method(args)
      if (lookahead() == :"!"){
	matchAny();
	Tree id = identifier();
	if (lookahead() == :"("){
	  //old matchAny();
	  Tree args = argumentList();
	  //old match(:")");
	  return new Tree(:objInvoke, tree, id, args);
	} else {
	  return new Tree(:objField, tree, id);
	}
      } else {
	return original(tree);
      }
    }

    extend Tree primaryTop(){
      if (lookahead() == :newObj){
	matchAny();
	Tree id = identifier();
	//old match(:"(");
	Tree args = argumentList();
	//old match(:")");
	return new Tree(:newObj, id, args);
      } else {
	return original();
      }
    }
    
    extend void initMacroTable() {
      original();
      extendMacro(:invoke1, new SystemMixinInvoke1());
      defineMacro(:objInvoke, new ObjInvokeMacro());
      defineMacro(:objField, new ObjFieldMacro());
      defineMacro(:newObj, new NewObjMacro());
    }
  }
}



class SystemMixinInvoke1 extends Macro {

  public Tree call(Tree tree){
    checkArgsLength(tree, 2);

    Symbol methodName = tree.args()[0].idName();
    Hashtable invokeStyleMacroTable
      = (Hashtable)Dynamic.get(:invokeStyleMacroTable);
    InvokeStyleMacro func =
      (InvokeStyleMacro)invokeStyleMacroTable.get(methodName);

    if (Dynamic.get(:expandingSystemName) == null
	|| func != null ){
      return orig.call(tree);
    } else {
      Tree[] args1	= tree.args();
      String name	= args1[0].idName().toString();
      Tree args		= args1[1];
      return Util.invokeExpCode(`(id self), name, args);
    }
  }
}


class ObjInvokeMacro extends Macro {
  public Tree call(Tree tree){
    checkArgsLength(tree, 3);
    Tree[] args1	= tree.args();
    Tree obj		= args1[0];
    String name		= args1[1].idName().toString();
    Tree args		= args1[2];
    return Util.invokeExpCode(obj, name, args);
  }
}


class ObjFieldMacro extends Macro {
  public Tree call(Tree tree){
    checkArgsLength(tree, 2);
    Tree[] args		= tree.args();
    Tree obj		= args[0];
    String name		= args[1].idName().toString();
    if (Opts.sml){
      // F_foo.v
      return `(fieldLong (name ,(Util.fieldNameToClass(name))) (id v));
    } else {
      // ((F_foo)obj.findField(:"foo")).value
      return `(fieldExp
	       (cast (name ,(Util.fieldNameToClass(name)))
		(invokeExp ,obj
		 (id findField)
		 (argumentList
		  ,(Util.nameToSymbol(name)))))
	       (id value));
    }
  }
}


class NewObjMacro extends Macro {
  public Tree call(Tree tree){
    checkArgsLength(tree, 2);
    Tree[] args1	= tree.args();
    String name		= args1[0].idName().toString();
    Tree args		= args1[1];
    return `(invokeLong (name ,(Util.classNameToClass(name)))
	       (id generateAndInit) ,args);
  }
}


//-------------------------------------------------

class Util {
  
  static boolean useSymbol = true;
  
  static Tree nameToSymbol(String name){
    if (useSymbol){
      return `(symbol ,(new Identifier(Symbol.intern(name))));
    } else {
      return new LiteralTree(new LiteralToken(:string, name));
    }
  }
  
  
  static Tree invokeExpCode(Tree obj, String name, Tree args){
    if (Opts.sml){
      // M_foo.m.call(a, b);
      return
	`(invokeLong
	  (name ,(Util.methodNameToClass(name)) (id m))
	  (id call)
	  (argumentList ,@(args.args())));
    } else {
      // ((M_foo)obj.findMethod(:"M_foo")).call(
      //     Obj.calleeself, Obj.calleemb, Obj.callembi ,a ,b)
      return
	`(invokeExp
	  (cast (name ,(Util.methodNameToClass(name)))
	   (invokeExp ,obj
	    (id findMethod)
	    (argumentList ,(nameToSymbol(name)))))
	  (id call)
	  (argumentList
	   (fieldLong (name (id Obj)) (id calleeself))
	   (fieldLong (name (id Obj)) (id calleemb))
	   (fieldLong (name (id Obj)) (id calleembi))
	   ,@(args.args())
	   ));
    }
  }

  static Tree methodNameToClass(String name){
    return new Identifier(Symbol.intern("M_"+ name));
  }

  static Tree fieldNameToClass(String name){
    return new Identifier(Symbol.intern("F_"+ name));
  }

  static Tree classNameToClass(String name){
    return new Identifier(Symbol.intern("C_"+ name));
  }
  
  static Tree mixinNameToClass(String name){
    return new Identifier(Symbol.intern("SystemMixin_"+ name));
  }
  
  static String getPackageName(){
    Obj epp = (Obj)Dynamic.get(:epp);
    Tree compilerUnit = epp!tree;
    Tree pack = compilerUnit.args()[0];
    if (pack.tag() != :packageDeclaration) throw Epp.fatal(""+ pack);
    if (pack.args().length == 0){
      return "";
    } else {
      Tree[] packageNameElems = pack.args()[0].args();
      StringBuffer buf = new StringBuffer();
      for (int i = 0; i < packageNameElems.length; i++){
	buf.append(packageNameElems[i].idName().toString());
	buf.append('.');
      }
      return buf.toString();
    }
  }
}
