/*
 *
 * Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002 Yuuji ICHISUGI
 *
 * Permission to use, copy, modify and redistribution this software in
 * whole and in part, for evaluation or research purposes and without fee
 * is hereby granted provided this copyright notice.
 * See CopyrightAndLicensing.txt for licensing condition.
 */

/*

 This is a program to clear *.java and *.class files under eppout directory.

	Usage:	java epp.ClearEppout
		 or
		java epp.Epp -clear

1. It checks if the directory "eppout" exists. If not, quits execution.
2. It checks all files contained under the "eppout" directory.
 If it finds files which are neither *.java files generated by EPP
 nor *.class files, quits execution.
 (*.java files are examined if it starts with "/* Generated by EPP".)
3. Then, it removes all files and sub-directories under the "eppout" directory.
 The "eppout" directory itself will remains.

*/

package epp;
import java.io.*;
import java.util.Vector;

public class ClearEppout {
  public static void main(String[] argv){
    new ClearEppout().start();
  }
  
  static final boolean debug = false;
  
  void start(){
    {
      File f = new File("eppout");
      if (! f.exists()){
	System.err.println("The directory \"eppout\" is not exists.");
	System.exit(1);
	throw new Error("never reached");
      }
    }

    Vector notGenerated = new Vector();
    Opts.out.println("Checking files under \"eppout\"...");
    
    checkRecursively("eppout", notGenerated);
    
    if (notGenerated.size() != 0){
      String[] names = new String[notGenerated.size()];
      notGenerated.copyInto(names);
      int printMax = 10;

      System.err.println();
      System.err.println("The followings are not generated files.");
      for (int i = 0; i < names.length && i < printMax ; i++){
	System.err.println(names[i]);
      }
      if (names.length >= printMax){
	System.err.println("...");
	System.err.println("Total "+ names.length+ " files.");
      }
      System.err.println();
      System.err.println("No files were removed. "+
			 "Please remove them manually.");
      System.exit(1);
      throw new Error("never reached");
    }
    
    // Remove all files under eppout.
    Opts.out.println("Start removing files under eppout...");
    String[] list = new File("eppout").list();
    for (int i = 0; i < list.length; i++){
      removeRecursively("eppout"+ File.separator+ list[i]);
    }
    
    Opts.out.println("Done.");
  }
  
  void checkRecursively(String path, Vector notGenerated){
    File f = new File(path);
    if (f.isDirectory()){
      String[] list = f.list();
      for (int i = 0; i < list.length; i++){
	checkRecursively(path+ File.separator+ list[i], notGenerated);
      }
    } else {
      if (! Epp.isGeneratedFile(path)){
	notGenerated.addElement(path);
      }
    }
  }
  
  void removeRecursively(String path){
    File f = new File(path);
    if (f.isDirectory()){
      String[] list = f.list();
      for (int i = 0; i < list.length; i++){
	removeRecursively(path+ File.separator+ list[i]);
      }
    } else {
      // Check it again.
      if (! Epp.isGeneratedFile(path)) {
	throw new Error("FATAL: File "+ path+ " is not a generated file.");
      }
    }
    if (debug) Opts.out.println("delete "+ path);
    f.delete();
  }
}
